﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.OptimizationExpertSystem.Common {
  [Item("K-Nearest Neighbor Recommender", "")]
  [StorableClass]
  public sealed class KNearestNeighborRecommender : ParameterizedNamedItem, IAlgorithmInstanceRecommender {
    private KnowledgeCenter okc;

    private IFixedValueParameter<EnumValue<ProblemInstanceProximityType>> ProximityTypeParameter {
      get { return (IFixedValueParameter<EnumValue<ProblemInstanceProximityType>>)Parameters["ProximityType"]; }
    }

    public ProblemInstanceProximityType ProximityType {
      get { return ProximityTypeParameter.Value.Value; }
      set { ProximityTypeParameter.Value.Value = value; }
    }

    private IFixedValueParameter<IntValue> KParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters["K"]; }
    }

    [StorableConstructor]
    private KNearestNeighborRecommender(bool deserializing) : base(deserializing) { }
    private KNearestNeighborRecommender(KNearestNeighborRecommender original, Cloner cloner)
      : base(original, cloner) { }
    public KNearestNeighborRecommender(KnowledgeCenter okc) {
      this.okc = okc;
      Parameters.Add(new FixedValueParameter<EnumValue<ProblemInstanceProximityType>>("ProximityType", "The type of neighbor proximity.", new EnumValue<ProblemInstanceProximityType>(ProblemInstanceProximityType.FeatureSpace)));
      Parameters.Add(new FixedValueParameter<IntValue>("K", "The number of nearest neighbors to consider.", new IntValue(5)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new KNearestNeighborRecommender(this, cloner);
    }

    public IEnumerable<IAlgorithm> GetRanking() {
      if (okc.Problem.ProblemId == -1) yield break;

      var distances = okc.GetProblemDistances(ProximityType);
      var K = KParameter.Value.Value;
      var performances = new Dictionary<IAlgorithm, List<double>>();
      for (var k = 0; k < K; k++) {
        if (distances.Count == 0) break;
        var min = distances.MinItems(x => x.Value).First();
        // lookup algorithm performances in min
        var perfs = okc.GetAlgorithmPerformance(min.Key);
        if (perfs.Count == 0) {
          k--;
          continue;
        }
        foreach (var p in perfs) {
          var ert = p.Value;
          if (double.IsNaN(ert)) ert = int.MaxValue;
          List<double> erts;
          if (!performances.TryGetValue(p.Key, out erts)) {
            performances[p.Key] = new List<double>() { ert }; ;
          } else erts.Add(ert);
        }
        distances.Remove(min.Key);
      }
      foreach (var alg in performances.Select(x => new { Alg = x.Key, Perf = x.Value.Average() })
                         .OrderBy(x => x.Perf)
                         .Select(x => x.Alg))
        yield return alg;
    }
  }
}
